import { MethodMap, MethodRegistry } from '../abstract-registry';
import {
  Argv,
  Awaitable,
  BeforeEventMap,
  Context,
  Events,
  I18n,
  MaybeArray,
  Middleware,
} from 'koishi';
import {
  BeforeEventName,
  BeforeEventNameAndPrepend,
  CommandRegisterConfig,
  EventName,
  EventNameAndPrepend,
  KoaContext,
  KoishiCommandDefinition,
  KoishiDoRegister,
  KoishiDoRegisterKeys,
  KoishiRouteDef,
  MappingStruct,
  PickEventFunction,
  PluginDefinition,
  TypedMethodDecorator,
} from '../../def';
import { Metadata } from '../../meta/metadata.decorators';
import { reflector } from '../../meta/meta-fetch';
import { CommandPut } from './command-put';
import { applySelector, generateRenderer } from '../../utility';
import { Next as KoaNext } from 'koa';
import { IncomingMessage } from 'http';

// eslint-disable-next-line @typescript-eslint/no-namespace
export namespace DoRegister {
  export function decorate<K extends keyof ConfigMap>(
    config: Config<K>,
  ): K extends keyof DecoratorMap
    ? TypedMethodDecorator<DecoratorMap[K]>
    : MethodDecorator {
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore
    return Metadata.set(KoishiDoRegister, config, KoishiDoRegisterKeys);
  }

  class SpecificRegistry extends MethodRegistry<
    ConfigMap,
    any,
    [
      Context,
      // obj
      any,
      // key
      string,
      // extra obj
      any | undefined,
      // view
      any | undefined,
    ]
  > {
    define<K extends keyof ConfigMap>(
      name: K,
      method: MethodMap<ConfigMap, any, [Context, any, string, ...any[]]>[K],
    ) {
      this.extend(name, method);
      return (config: ConfigMap[K]) =>
        decorate<K>({ type: name, data: config });
    }
  }

  export interface ConfigMap {
    middleware: boolean; // prepend
    onEvent: EventNameAndPrepend;
    beforeEvent: BeforeEventNameAndPrepend;
    plugin: void;
    command: CommandRegisterConfig;
    route: KoishiRouteDef;
    ws: MaybeArray<string | RegExp>;
    formatter: string;
    preset: string;
    interval: number;
  }

  export interface DecoratorMap {
    middleware: Middleware;
    onEvent: PickEventFunction<Events, EventName>;
    beforeEvent: PickEventFunction<BeforeEventMap, BeforeEventName>;
    plugin(): Awaitable<PluginDefinition | undefined>;
    command(...args: any[]): Awaitable<string | void | undefined>;
    route(ctx: KoaContext, Next: KoaNext): any;
    ws(socket: WebSocket, request: IncomingMessage): any;
    formatter: I18n.Formatter;
    preset: I18n.Renderer;
  }

  export type Config<K extends keyof ConfigMap = keyof ConfigMap> =
    MappingStruct<ConfigMap, K>;

  export const registry = new SpecificRegistry();

  export const middleware = registry.define(
    'middleware',
    (prepend, ctx, obj, key) =>
      ctx.middleware((session, next) => obj[key](session, next), prepend),
  );

  export const event = registry.define('onEvent', (data, ctx, obj, key) =>
    ctx.on(
      data.name,
      (session, ...args) => obj[key](session, ...args),
      data.prepend,
    ),
  );

  export const beforeEvent = registry.define(
    'beforeEvent',
    (data, ctx, obj, key) =>
      ctx.before(
        data.name,
        (session, ...args) => obj[key](session, ...args),
        data.prepend,
      ),
  );

  export const route = registry.define('route', (data, ctx, obj, key) => {
    const path = data.path.startsWith('/') ? data.path : `/${data.path}`;
    return ctx.router[data.method](path, (koaCtx, next) =>
      obj[key](koaCtx, next),
    );
  });

  export const ws = registry.define('ws', (data, ctx, obj, key) =>
    ctx.router.ws(data, (socket, request) => obj[key](socket, request)),
  );

  function applyInnerPluginDef(
    ctx: Context,
    key: string,
    pluginDef: PluginDefinition<any>,
  ) {
    const pluginCtx = applySelector(ctx, pluginDef);
    if (pluginDef == null) {
      return;
    }
    if (!pluginDef || !pluginDef.plugin) {
      throw new Error(`Invalid plugin from method ${key}.`);
    }
    pluginCtx.plugin(pluginDef.plugin, pluginDef.options);
  }

  export const plugin = registry.define('plugin', (_, ctx, obj, key) => {
    const pluginDescMayBeProm: Awaitable<PluginDefinition<any>> = obj[key]();
    if (pluginDescMayBeProm instanceof Promise) {
      return pluginDescMayBeProm.then((pluginDef) => {
        applyInnerPluginDef(ctx, key, pluginDef);
      });
    } else {
      applyInnerPluginDef(ctx, key, pluginDescMayBeProm);
      return;
    }
  });

  export const command = registry.define(
    'command',
    (data, ctx, obj, key, extraObj, view) => {
      let command = ctx.command(data.def, data.desc, data.config);
      const commandDefs = reflector.getProperty(
        KoishiCommandDefinition,
        obj,
        key,
        extraObj,
      );
      const renderer = generateRenderer(view);
      for (const commandDef of commandDefs) {
        command = commandDef(command, ctx, renderer) || command;
      }
      if (!data.config?.empty) {
        if (!data.putOptions) {
          command.action((argv: Argv, ...args: any[]) =>
            obj[key](argv, ...args),
          );
        } else {
          for (let i = 0; i < data.putOptions.length; i++) {
            const putOption = data.putOptions[i];
            if (!putOption) {
              continue;
            }
            const nativeType = data.paramTypes[i];
            CommandPut.preRegistry.execute(
              putOption,
              command,
              ctx,
              nativeType,
              view,
            );
          }
          command.action((argv: Argv, ...args: any[]) => {
            const params = data.putOptions.map((o, i) =>
              CommandPut.registry.execute(
                o,
                argv,
                args,
                data.paramTypes[i],
                view,
              ),
            );
            return obj[key](...params);
          });
        }
      }
      return command;
    },
  );

  export const formatter = registry.define(
    'formatter',
    (data, ctx, obj, key) => {
      ctx.i18n.formatter(data, (...args) => obj[key](...args));
      ctx.on('dispose', () => {
        delete ctx.i18n._formatters[data];
      });
      return;
    },
  );

  export const preset = registry.define('preset', (data, ctx, obj, key) => {
    ctx.i18n.preset(data, (...args) => obj[key](...args));
    ctx.on('dispose', () => {
      delete ctx.i18n._presets[data];
    });
    return;
  });

  export const interval = registry.define('interval', (data, ctx, obj, key) => {
    ctx.setInterval((...args) => obj[key](...args), data);
  });
}
